# © Copyright Databand.ai, an IBM Company 2022
from typing import Dict, List

from dbnd import parameter
from dbnd._core.task import Config


class AirflowTrackingConfig(Config):
    _conf__task_family = "airflow_tracking"

    spark_submit_dbnd_java_agent = parameter(
        default=None,
        description="A dbnd java agent jar which used to track a java application, located on the local machine",
    )[str]

    databricks_dbnd_java_agent = parameter(
        default=None,
        description="A dbnd java agent jar which used to track a java application, located on remote machine",
    )[str]

    track_airflow_execute_result = parameter(
        default=False,
        description="Allow saving the results of tracked airflow operator",
    )[bool]

    track_xcom_values = parameter(
        default=False,
        description="Allow logging the values of xcom variables from airflow",
    )[bool]

    max_xcom_length = parameter(
        default=10, description="The amount of xcom values to track, per operator"
    )[int]

    # This shouldn't be set to None. In webserver None evaluates to True.
    af_with_monitor = parameter(
        default=True, description="Activate when airflow monitor is not in use"
    )[bool]

    sql_reporting = parameter(
        default=False, description="Enable targets reporting from sql queries"
    )[bool]


class TrackingSparkConfig(Config):
    _conf__task_family = "tracking_spark"

    provide_databand_service_endpoint = parameter(
        default=True,
        description="Should Databand inject tracker URL and access token into spark-submit cmd,"
        "e.g. `spark-submit --conf DBND__CORE__DATABAND_URL=... --conf DBND__CORE__DATABAND_ACCESS_TOKEN=... script.py",
    )[bool]

    agent_path = parameter(
        default=None,
        description="Path to Databand Agent jar to be added to the spark job as java agent. Jar file has to be placed directly to the cluster filesystem.",
    )[str]

    jar_path = parameter(
        default=None,
        description="Path to Databand Agent jar to be added to the spark job as regular jar file. Jar file can be placed to the S3/GCS/DBFS as well as directly to the cluster filesystem.",
    )[str]

    query_listener = parameter(
        default=False,
        description="Should Databand turn on Spark Query Listener and automatically collect datasource operations",
    )[bool]

    def spark_conf(self):
        result = {}
        if self.agent_path:
            result["spark.driver.extraJavaOptions"] = "-javaagent:{agent_path}".format(
                agent_path=self.agent_path
            )
        if self.jar_path:
            result["spark.jars"] = self.jar_path
        if self.query_listener:
            result[
                "spark.sql.queryExecutionListeners"
            ] = "ai.databand.spark.DbndSparkQueryExecutionListener"
        return result

    def merged_spark_conf(self, origin_spark_conf):
        # type: (Dict[str, str])->Dict[str, str]
        """
        Operator spark conf can contain "spark.jars", "spark.driver.extraJavaOptions"
        and "spark.sql.queryExecutionListeners".
        To preserve existing properties we need to concat them with patched properties
        """
        new_spark_conf = self.spark_conf()
        return self.concat_properties(
            origin_spark_conf,
            new_spark_conf,
            [
                "spark.driver.extraJavaOptions",
                "spark.jars",
                "spark.sql.queryExecutionListeners",
            ],
        )

    def concat_properties(self, origin_conf, new_conf, property_names):
        # type: (Dict[str, str], Dict[str, str], List[str])-> Dict[str, str]
        result = {}
        for property_name in property_names:
            if property_name in origin_conf and property_name in new_conf:
                result[property_name] = (
                    origin_conf[property_name] + "," + new_conf[property_name]
                )
            elif property_name not in origin_conf and property_name in new_conf:
                result[property_name] = new_conf[property_name]
        return result



# © Copyright Databand.ai, an IBM Company 2022

import logging

import dbnd


AIRFLOW_LEGACY_URL_KEY = "airflow"
logger = logging.getLogger(__name__)


@dbnd.hookimpl
def dbnd_setup_plugin():
    from dbnd import register_config_cls
    from dbnd_airflow.config import AirflowConfig

    register_config_cls(AirflowConfig)



# © Copyright Databand.ai, an IBM Company 2022

import getpass

from airflow import DAG, models, settings

from dbnd_airflow.constants import AIRFLOW_VERSION_2


def safe_get_context_manager_dag():
    """
    Try to find the CONTEXT_MANAGER_DAG object inside airflow.
    It was moved between versions, so we look for it in all the hiding places that we know of.
    """
    if AIRFLOW_VERSION_2:
        from airflow.models.dag import DagContext

        return DagContext.get_current_dag()

    if hasattr(settings, "CONTEXT_MANAGER_DAG"):
        return settings.CONTEXT_MANAGER_DAG
    elif hasattr(DAG, "_CONTEXT_MANAGER_DAG"):
        return DAG._CONTEXT_MANAGER_DAG
    elif hasattr(models, "_CONTEXT_MANAGER_DAG"):
        return models._CONTEXT_MANAGER_DAG

    return None


def hostname_as_username():
    """
    This function is used to speed up airflow.
    When not redirected to this method, airflow retrieves the host name of the machine using the `socket` library.
    The calls to `socket` are extremely slow, so we redirect it to the native python implementation
    """
    return getpass.getuser()


def is_airflow_support_template_fields():
    from airflow import __version__

    return __version__ not in {"1.10.0", "1.10.1"}


def DbndFunctionalOperator_1_10_0(
    task_id, dbnd_task_type, dbnd_task_params_fields, **ctor_kwargs
):
    """
    Workaround for backwards compatibility with Airflow 1.10.0,
    dynamically creating new operator class (that inherits AirflowDagDbndOperator).
    This is used because XCom templates for each operator are found in a class attribute
    and we have different XCom templates for each INSTANCE of the operator.
    Also, the template_fields attribute has a different value for each task, so we must ensure that they
    don't get mixed up.
    """
    template_fields = dbnd_task_params_fields
    from dbnd_airflow.functional.dbnd_functional_operator import DbndFunctionalOperator

    new_op = type(
        dbnd_task_type, (DbndFunctionalOperator,), {"template_fields": template_fields}
    )
    op = new_op(
        task_id=task_id,
        dbnd_task_type=dbnd_task_type,
        dbnd_task_params_fields=dbnd_task_params_fields,
        **ctor_kwargs
    )
    return op



# © Copyright Databand.ai, an IBM Company 2022

import io

from typing import List, Tuple

from dbnd_airflow.airflow_utils import safe_get_context_manager_dag
from targets import AtomicLocalFile
from targets.fs import register_file_system
from targets.fs.file_system import FileSystem
from targets.pipes.base import FileWrapper


def build_xcom_str(op, name=None):
    """
    :param op: airflow operator
    :param name: name of result
    :return:
    """
    result_key = "['%s']" % name if name else ""
    xcom_path = "{{task_instance.xcom_pull('%s')%s}}" % (op.task_id, result_key)
    return XComStr(xcom_path, op.task_id)


class _BaseOperator(object):
    """
    Abstract class implementing down and up stream operations.
    """

    @property
    def op(self):
        raise NotImplemented()

    def set_downstream(self, task_or_task_list):
        self.op.set_downstream(task_or_task_list)

    def set_upstream(self, task_or_task_list):
        self.op.set_upstream(task_or_task_list)

    def __lshift__(self, other):
        return self.set_upstream(other)

    def __rshift__(self, other):
        return self.set_downstream(other)


class XComStr(str, _BaseOperator):
    def __new__(cls, value, task_id):
        # explicitly only pass value to the str constructor
        obj = super(XComStr, cls).__new__(cls, value)
        obj.task_id = task_id
        return obj

    def __iter__(self):
        raise TypeError(
            "It seems that you are trying to assign output to multiple values. If your task's function "
            "returns multiple values please add type annotations to it."
        )

    @property
    def op(self):
        return safe_get_context_manager_dag().get_task(self.task_id)


class XComResults(_BaseOperator):
    target_no_traverse = True

    def __init__(self, result, sub_results):
        # type: (XComStr, List[ Tuple[str, XComStr]]) -> XComResults
        self.xcom_result = result
        self.sub_results = sub_results
        self._names = [n for n, _ in sub_results]
        for k, v in self.sub_results:
            if not hasattr(self, k):
                setattr(self, k, v)

    @property
    def names(self):
        return self._names

    def __iter__(self):
        return iter(xcom_sub_result for _, xcom_sub_result in self.sub_results)

    def __repr__(self):
        return "result(%s)" % ",".join(self.names)

    def as_dict(self):
        return dict(self.sub_results)

    @property
    def op(self):
        return safe_get_context_manager_dag().get_task(self.sub_results[0][1].task_id)

    def __getitem__(self, key):
        return self.as_dict()[key]


class AirflowXComFileSystem(FileSystem):
    def exists(self, path):
        return False

    def open_read(self, path, mode="r"):
        return FileWrapper(io.BufferedReader(io.FileIO(path, mode)))

    def open_write(self, path, mode="w", **kwargs):
        return AtomicLocalFile(path, fs=self, mode=mode)


class XComAtomicFile(AtomicLocalFile):
    def move_to_final_destination(self):
        self.fs.move_from_local(self.tmp_path, self.path)


register_file_system("xcom", AirflowXComFileSystem)
register_file_system("jinja", AirflowXComFileSystem)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_airflow.constants import AIRFLOW_VERSION_1, AIRFLOW_VERSION_2


if AIRFLOW_VERSION_1:
    from airflow.hooks.base_hook import BaseHook

elif AIRFLOW_VERSION_2:
    from airflow.hooks.base import BaseHook

else:
    raise NotImplementedError

__all__ = ["BaseHook"]
